import os

from babel.messages import pofile, mofile


def load_po(filename, **kwargs):
    """read po/pot file and return catalog object

    :param unicode filename: path to po/pot file
    :param kwargs: keyword arguments to forward to babel's read_po call
    :return: catalog object
    """
    # pre-read to get charset
    with open(filename, "rb") as f:
        cat = pofile.read_po(f)
    charset = cat.charset or "utf-8"

    # To decode lines by babel, read po file as binary mode and specify charset for
    # read_po function.
    with open(filename, "rb") as f:  # FIXME: encoding VS charset
        return pofile.read_po(f, charset=charset, **kwargs)


def dump_po(filename, catalog, **kwargs):
    """write po/pot file from catalog object

    :param unicode filename: path to po file
    :param catalog: catalog object
    :param kwargs: keyword arguments to forward to babel's write_po call; also
                    accepts a deprecated `line_width` option to forward to
                    write_po's `width` option
    :return: None
    """
    dirname = os.path.dirname(filename)
    os.makedirs(dirname, exist_ok=True)

    # (compatibility) line_width was the original argument used to forward
    # line width hints into write_po's `width` argument; if provided,
    # set/override the width value
    if "line_width" in kwargs:
        kwargs["width"] = kwargs["line_width"]
        del kwargs["line_width"]

    # Because babel automatically encode strings, file should be open as binary mode.
    with open(filename, "wb") as f:
        pofile.write_po(f, catalog, **kwargs)


def write_mo(filename, catalog, **kwargs):
    """write mo file from catalog object

    :param unicode filename: path to mo file
    :param catalog: catalog object
    :return: None
    """
    dirname = os.path.dirname(filename)
    os.makedirs(dirname, exist_ok=True)

    with open(filename, "wb") as f:
        mofile.write_mo(f, catalog, **kwargs)


def translated_entries(catalog):
    return [m for m in catalog if m.id and m.string]


def fuzzy_entries(catalog):
    return [m for m in catalog if m.id and m.fuzzy]


def untranslated_entries(catalog):
    return [m for m in catalog if m.id and not m.string]


def update_with_fuzzy(catalog, catalog_source):
    """update catalog by template catalog with fuzzy flag.

    :param catalog: catalog object to be updated
    :param catalog_source: catalog object as a template to update 'catalog'
    :return: None
    """
    catalog.update(catalog_source)



"""
Python compatibility functions.
"""

import sys
import os
import warnings
from typing import Any, Callable

# ------------------------------------------------------------------------------
# Python 2/3 compatibility
fs_encoding = sys.getfilesystemencoding() or sys.getdefaultencoding()


def relpath(path: str, start: str = os.curdir) -> str:
    """Return a relative version of a path"""
    try:
        return os.path.relpath(path, start)
    except ValueError:
        return path


# convert_with_2to3():
# support for running 2to3 over config files
def convert_with_2to3(filepath: str) -> str:
    from lib2to3.refactor import RefactoringTool, get_fixers_from_package
    from lib2to3.pgen2.parse import ParseError

    fixers = get_fixers_from_package("lib2to3.fixes")
    refactoring_tool = RefactoringTool(fixers)
    source = refactoring_tool._read_python_source(filepath)[0]
    try:
        tree = refactoring_tool.refactor_string(source, "conf.py")
    except ParseError as err:
        # do not propagate lib2to3 exceptions
        lineno, offset = err.context[1]
        # try to match ParseError details with SyntaxError details
        raise SyntaxError(err.msg, (filepath, lineno, offset, err.value))
    return str(tree)


def execfile_(filepath: str, _globals: Any, open: Callable = open) -> None:
    with open(filepath, "rb") as f:
        source = f.read()

    # compile to a code object, handle syntax errors
    filepath_enc = filepath.encode(fs_encoding)
    try:
        code = compile(source, filepath_enc, "exec")
    except SyntaxError:
        # maybe the file uses 2.x syntax; try to refactor to
        # 3.x syntax using 2to3
        source = convert_with_2to3(filepath)
        code = compile(source, filepath_enc, "exec")
        warnings.warn(
            "Support for evaluating Python 2 syntax is deprecated "
            "and will be removed in sphinx-intl 4.0. "
            "Convert %s to Python 3 syntax.",
            source=filepath,
        )
    exec(code, _globals)



import dataclasses
import multiprocessing as mp
import os
from glob import glob
from typing import Optional

import click

from . import catalog as c
from .pycompat import relpath


# ==================================
# utility functions


def get_lang_dirs(path):
    dirs = [
        relpath(d, path)
        for d in glob(path + "/[a-z]*")
        if os.path.isdir(d) and not d.endswith("pot")
    ]
    return (tuple(dirs),)


# ==================================
# commands


@dataclasses.dataclass(frozen=True)
class UpdateItem:
    po_file: str
    pot_file: str
    lang: str
    line_width: int
    ignore_obsolete: bool


@dataclasses.dataclass(frozen=True)
class UpdateResult:
    po_file: str
    status: str
    added: Optional[int] = 0
    deleted: Optional[int] = 0


def _update_single_file(update_item: UpdateItem):
    cat_pot = c.load_po(update_item.pot_file)
    if os.path.exists(update_item.po_file):
        cat = c.load_po(update_item.po_file)
        msgids = {m.id for m in cat if m.id}
        c.update_with_fuzzy(cat, cat_pot)
        new_msgids = {m.id for m in cat if m.id}
        if msgids != new_msgids:
            added = new_msgids - msgids
            deleted = msgids - new_msgids
            c.dump_po(
                update_item.po_file,
                cat,
                width=update_item.line_width,
                ignore_obsolete=update_item.ignore_obsolete,
            )
            return UpdateResult(update_item.po_file, "update", len(added), len(deleted))
        else:
            return UpdateResult(update_item.po_file, "notchanged")
    else:  # new po file
        cat_pot.locale = update_item.lang
        c.dump_po(
            update_item.po_file,
            cat_pot,
            width=update_item.line_width,
            ignore_obsolete=update_item.ignore_obsolete,
        )
        return UpdateResult(update_item.po_file, "create")


def update(
    locale_dir, pot_dir, languages, line_width=76, ignore_obsolete=False, jobs=0
):
    """
    Update specified language's po files from pot.

    :param unicode locale_dir: path for locale directory
    :param unicode pot_dir: path for pot directory
    :param tuple languages: languages to update po files
    :param number line_width: maximum line width of po files
    :param bool ignore_obsolete: ignore obsolete entries in po files
    :param number jobs: number of CPUs to use
    :return: {'create': 0, 'update': 0, 'notchanged': 0}
    :rtype: dict
    """
    status = {
        "create": 0,
        "update": 0,
        "notchanged": 0,
    }

    to_translate = []
    for dirpath, dirnames, filenames in os.walk(pot_dir):
        for filename in filenames:
            pot_file = os.path.join(dirpath, filename)
            base, ext = os.path.splitext(pot_file)
            if ext != ".pot":
                continue
            basename = relpath(base, pot_dir)
            for lang in languages:
                po_dir = os.path.join(locale_dir, lang, "LC_MESSAGES")
                po_file = os.path.join(po_dir, basename + ".po")
                to_translate.append(
                    UpdateItem(po_file, pot_file, lang, line_width, ignore_obsolete)
                )

    with mp.Pool(processes=jobs or None) as pool:
        for result in pool.imap_unordered(_update_single_file, to_translate):
            status[result.status] += 1
            if result.status == "update":
                click.echo(
                    f"Update: {result.po_file} +{result.added}, -{result.deleted}"
                )
            elif result.status == "create":
                click.echo(f"Create: {result.po_file}")
            else:
                click.echo(f"Not Changed: {result.po_file}")

    return status


def build(locale_dir, output_dir, languages):
    """
    Build specified language's po files into mo.

    :param unicode locale_dir: path for locale directory
    :param unicode output_dir: path for mo output directory
    :param tuple languages: languages to update po files
    :return: None
    """
    for lang in languages:
        lang_dir = os.path.join(locale_dir, lang)
        for dirpath, dirnames, filenames in os.walk(lang_dir):
            dirpath_output = os.path.join(
                output_dir, os.path.relpath(dirpath, locale_dir)
            )

            for filename in filenames:
                base, ext = os.path.splitext(filename)
                if ext != ".po":
                    continue

                mo_file = os.path.join(dirpath_output, base + ".mo")
                po_file = os.path.join(dirpath, filename)

                if os.path.exists(mo_file) and os.path.getmtime(
                    mo_file
                ) > os.path.getmtime(po_file):
                    continue
                click.echo(f"Build: {mo_file}")
                cat = c.load_po(po_file)
                c.write_mo(mo_file, cat)


def stat(locale_dir, languages):
    """
    Print statistics for all po files.

    :param unicode locale_dir: path for locale directory
    :param tuple languages: languages to update po files
    :return: {'FILENAME': {'translated': 0, 'fuzzy': 0, 'untranslated': 0}, ...}
    :rtype: dict
    """
    result = {}

    for lang in languages:
        lang_dir = os.path.join(locale_dir, lang)
        for dirpath, dirnames, filenames in os.walk(lang_dir):
            for filename in filenames:
                po_file = os.path.join(dirpath, filename)
                base, ext = os.path.splitext(po_file)
                if ext != ".po":
                    continue

                cat = c.load_po(po_file)
                r = result[po_file.replace("\\", "/")] = {
                    "translated": len(c.translated_entries(cat)),
                    "fuzzy": len(c.fuzzy_entries(cat)),
                    "untranslated": len(c.untranslated_entries(cat)),
                }
                click.echo(
                    "{}: {} translated, {} fuzzy, {} untranslated.".format(
                        po_file,
                        r["translated"],
                        r["fuzzy"],
                        r["untranslated"],
                    )
                )

    return result



import os
import re
import subprocess
import textwrap
from pathlib import Path
from shutil import which

import click

from .catalog import load_po


# ==================================
# settings

TRANSIFEX_CLI_MINIMUM_VERSION = (1, 2, 1)

# To avoid using invalid resource name, append underscore to such names.
# As a limitation, append `_` doesn't care about collision to other resources.
# e.g. 'glossary' and 'glossary_' are pushed as a 'glossary_'. The following
# resource names are reserved slugs, Transifex will reply with an error on these
# resource names.
IGNORED_RESOURCE_NAMES = (
    "glossary",
    "settings",
)

TRANSIFEXRC_TEMPLATE = """\
[https://www.transifex.com]
rest_hostname = https://rest.api.transifex.com
token = %(transifex_token)s
"""

TXCONFIG_TEMPLATE = """\
[main]
host = https://www.transifex.com
"""


# ==================================
# utility functions


def normalize_resource_name(name):
    # replace path separator with '--'
    name = re.sub(r"[\\/]", "--", name)

    # replace unusable characters (not: -, _ ascii, digit) with '_'
    name = re.sub(r"[^\-\w]", "_", name)

    # append `_` for ignored resource names
    while name in IGNORED_RESOURCE_NAMES:
        name += "_"

    return name


def check_transifex_cli_installed():
    tx_cli_url = "https://raw.githubusercontent.com/transifex/cli/master/install.sh"
    if not which("tx"):
        msg = textwrap.dedent(f"""\
            Could not run "tx".
            You need to install the Transifex CLI external library.
            Please install the below command and restart your terminal \
            if you want to use this action:

                $ curl -o- {tx_cli_url} | bash

            """)
        raise click.BadParameter(msg)

    version_msg = subprocess.check_output("tx --version", shell=True, encoding="utf-8")

    if not version_msg.startswith("TX Client"):
        msg = textwrap.dedent(f"""\
            The old transifex_client library was found.
            You need to install the Transifex CLI external library.
            Please install the below command and restart your terminal \
            if you want to use this action:

                $ curl -o- {tx_cli_url} | bash

            """)
        raise click.BadParameter(msg)

    version = tuple(int(x) for x in version_msg.split("=")[-1].strip().split("."))

    if not version >= TRANSIFEX_CLI_MINIMUM_VERSION:
        msg = textwrap.dedent(f"""\
        An unsupported version of the Transifex CLI was found.
        Version {TRANSIFEX_CLI_MINIMUM_VERSION} or greater is required.
        Please run the below command if you want to use this action:

            $ tx update

        """)
        raise click.BadParameter(msg)


# ==================================
# commands


def create_transifexrc(transifex_token):
    """
    Create `$HOME/.transifexrc`
    """
    target = os.path.normpath(os.path.expanduser("~/.transifexrc"))

    if os.path.exists(target):
        click.echo(f"{target} already exists, skipped.")
        return

    if not transifex_token:
        msg = textwrap.dedent("""\
        You need a transifex token by command option or environment.
        command option: --transifex-token
        """)
        raise click.BadParameter(msg, param_hint="transifex_token")

    with open(target, "w") as rc:
        rc.write(TRANSIFEXRC_TEMPLATE % locals())
    click.echo(f"Create: {target}")


def create_txconfig():
    """
    Create `./.tx/config`
    """
    target = os.path.normpath(".tx/config")
    if os.path.exists(target):
        click.echo(f"{target} already exists, skipped.")
        return

    if not os.path.exists(".tx"):
        os.mkdir(".tx")

    with open(target, "w") as f:
        f.write(TXCONFIG_TEMPLATE)

    click.echo(f"Create: {target}")


def update_txconfig_resources(
    transifex_organization_name, transifex_project_name, locale_dir, pot_dir
):
    """
    Update resource sections of `./.tx/config`.
    """
    check_transifex_cli_installed()

    cmd_tmpl = (
        "tx",
        "add",
        "--organization",
        "%(transifex_organization_name)s",
        "--project",
        "%(transifex_project_name)s",
        "--resource",
        "%(resource_slug)s",
        "--resource-name",
        "%(resource_name)s",
        "--file-filter",
        "%(locale_dir)s/<lang>/LC_MESSAGES/%(resource_path)s.po",
        "--type",
        "PO",
        "%(pot_dir)s/%(resource_path)s.pot",
    )

    # convert transifex_project_name to internal name
    transifex_project_name = transifex_project_name.replace(" ", "-")
    transifex_project_name = re.sub(r"[^\-_\w]", "", transifex_project_name)

    pot_dir = Path(pot_dir)
    pot_paths = sorted(pot_dir.glob("**/*.pot"))
    with click.progressbar(
        pot_paths,
        length=len(pot_paths),
        color="green",
        label="adding pots...",
        item_show_func=lambda p: str(p),
    ) as progress_bar:
        for pot_path in progress_bar:
            resource_path = str(pot_path.relative_to(pot_dir).with_suffix(""))
            resource_slug = resource_name = normalize_resource_name(resource_path)
            pot = load_po(str(pot_path))
            if len(pot):
                lv = locals()
                cmd = [arg % lv for arg in cmd_tmpl]
                subprocess.check_output(cmd, shell=False)
            else:
                click.echo(f"{pot_path} is empty, skipped")



"""
sphinx-intl
~~~~~~~~~~~
Sphinx utility that make it easy to translate and to apply translation.

:copyright: Copyright 2019 by Takayuki SHIMIZUKAWA.
:license: BSD, see LICENSE for details.
"""

import re
import os
import warnings
from glob import glob

import click
from sphinx.util.tags import Tags

from . import basic
from . import transifex
from .pycompat import execfile_, relpath

ENVVAR_PREFIX = "SPHINXINTL"


# ==================================
# utility functions


def read_config(path, passed_tags):
    tags = Tags()
    passed_tags = sum(passed_tags, ())
    for tag in passed_tags:
        tags.add(tag)

    namespace = {
        "__file__": os.path.abspath(path),
        "tags": tags,
    }

    olddir = os.getcwd()
    try:
        if not os.path.isfile(path):
            msg = "'%s' is not found (or specify --locale-dir option)." % path
            raise click.BadParameter(msg)
        os.chdir(os.path.dirname(path) or ".")
        execfile_(os.path.basename(path), namespace)
    finally:
        os.chdir(olddir)

    return namespace


def get_lang_dirs(path):
    dirs = [
        relpath(d, path)
        for d in glob(path + "/[a-z]*")
        if os.path.isdir(d) and not d.endswith("pot")
    ]
    return (tuple(dirs),)


# ==================================
# click options


class LanguagesType(click.ParamType):
    name = "languages"
    envvar_list_splitter = ","

    def convert(self, value, param, ctx):
        langs = value.split(",")
        return tuple(langs)


LANGUAGES = LanguagesType()


class TagsType(click.ParamType):
    name = "tags"
    envvar_list_splitter = ","

    def convert(self, value, param, ctx):
        tags = value.split(",")
        return tuple(tags)


TAGS = TagsType()


option_locale_dir = click.option(
    "-d",
    "--locale-dir",
    envvar=ENVVAR_PREFIX + "_LOCALE_DIR",
    type=click.Path(exists=False, file_okay=False),
    default="locales",
    metavar="<DIR>",
    show_default=True,
    help="locale directories that allow comma separated string. This option "
    "override locale_dir in conf.py setting if provided. Default is empty "
    "list.",
)

option_pot_dir = click.option(
    "--pot-dir",
    "-p",
    envvar=ENVVAR_PREFIX + "_POT_DIR",
    type=click.Path(exists=False, file_okay=False),
    metavar="<DIR>",
    show_default=True,
    help="pot files directory which is generated by sphinx. "
    "Default is 'pot' directory under '--locale-dir' path.",
)

option_output_dir = click.option(
    "--output-dir",
    "-o",
    envvar=ENVVAR_PREFIX + "_OUTPUT_DIR",
    type=click.Path(exists=False, file_okay=False),
    metavar="<DIR>",
    show_default=True,
    help="mo files directory where files are written. "
    "Default is to match the '--locale-dir' path.",
)

option_tag = click.option(
    "-t",
    "--tag",
    envvar=ENVVAR_PREFIX + "_TAG",
    type=TAGS,
    metavar="<TAG>",
    show_default=True,
    multiple=True,
    help="Pass tags to conf.py, as same as passed to sphinx-build -t option.",
)

option_language = click.option(
    "-l",
    "--language",
    envvar=ENVVAR_PREFIX + "_LANGUAGE",
    type=LANGUAGES,
    metavar="<LANG>",
    show_default=True,
    multiple=True,
    help="Target language to update po files. Default is ALL.",
)

option_line_width = click.option(
    "-w",
    "--line-width",
    envvar=ENVVAR_PREFIX + "_LINE_WIDTH",
    type=int,
    default=76,
    metavar="<WIDTH>",
    show_default=True,
    multiple=False,
    help="The maximum line width for the po files, 0 or a negative number "
    "disable line wrapping",
)

option_jobs = click.option(
    "-j",
    "--jobs",
    envvar=ENVVAR_PREFIX + "_JOBS",
    type=int,
    default=0,
    metavar="<JOBS>",
    show_default=True,
    multiple=False,
    help="The number of CPUs to use for updates. 0 means all",
)

option_no_obsolete = click.option(
    "--no-obsolete",
    envvar=ENVVAR_PREFIX + "_NO_OBSOLETE",
    is_flag=True,
    default=False,
    help="Remove obsolete #~ messages.",
)

option_transifex_token = click.option(
    "--transifex-token",
    envvar=ENVVAR_PREFIX + "_TRANSIFEX_TOKEN",
    type=str,
    metavar="<TOKEN>",
    required=True,
    help="Your transifex token. (DEPRECATED)",
)

option_transifex_organization_name = click.option(
    "--transifex-organization-name",
    envvar=ENVVAR_PREFIX + "_TRANSIFEX_ORGANIZATION_NAME",
    type=str,
    metavar="<ORGANIZATION-NAME>",
    required=True,
    help="Your transifex organization name.",
)

option_transifex_project_name = click.option(
    "--transifex-project-name",
    envvar=ENVVAR_PREFIX + "_TRANSIFEX_PROJECT_NAME",
    type=str,
    metavar="<PROJECT-NAME>",
    required=True,
    help="Your transifex project name.",
)


# ==================================
# commands


@click.group()
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    default=None,
    metavar="<FILE>",
    help="Sphinx conf.py file to read a locale directory setting.",
)
@option_tag
@click.pass_context
def main(ctx, config, tag):
    # load conf.py
    ctx.config = config
    if ctx.config is None:
        for c in ("conf.py", "source/conf.py"):
            if os.path.exists(c):
                ctx.config = c
                break

    # for locale_dir
    ctx.locale_dir = None
    if ctx.config:
        cfg = read_config(ctx.config, tag)
        # Use explicit locale_dirs if set, otherwise use Sphinx's default ['locales']
        locale_dirs = cfg.get("locale_dirs", ["locales"])
        ctx.locale_dir = os.path.join(os.path.dirname(ctx.config), locale_dirs[0])

    # for pot_dir
    ctx.pot_dir = None
    for d in ("_build/gettext", "build/gettext", "_build/locale", "build/locale"):
        if os.path.exists(d):
            ctx.pot_dir = d
            break

    # for transifex_project_name
    ctx.transifex_project_name = None
    target = os.path.normpath(".tx/config")
    if os.path.exists(target):
        matched = re.search(r"\[(.*)\..*\]", open(target).read())
        if matched:
            ctx.transifex_project_name = matched.groups()[0]
            click.echo(
                "Project name loaded from .tx/config: {}".format(
                    ctx.transifex_project_name
                )
            )

    ctx.default_map = {
        "update": {
            "locale_dir": ctx.locale_dir,
            "pot_dir": ctx.pot_dir,
        },
        "build": {
            "locale_dir": ctx.locale_dir,
        },
        "stat": {
            "locale_dir": ctx.locale_dir,
        },
        "update-txconfig-resources": {
            "locale_dir": ctx.locale_dir,
            "pot_dir": ctx.pot_dir,
            "transifex_project_name": ctx.transifex_project_name,
        },
    }


@main.command()
@option_locale_dir
@option_pot_dir
@option_language
@option_line_width
@option_no_obsolete
@option_jobs
def update(locale_dir, pot_dir, language, line_width, no_obsolete, jobs):
    """
    Update specified language's po files from pot.

    \b
    For examples:
       sphinx-intl update -l de -l ja
       sphinx-intl update -l de,ja
    """
    if not pot_dir:
        pot_dir = os.path.join(locale_dir, "pot")
    if not os.path.exists(pot_dir):
        msg = (
            "%(pot_dir)r does not exist. Please specify pot directory with "
            "-p option, or preparing your pot files in %(pot_dir)r." % locals()
        )
        raise click.BadParameter(msg, param_hint="pot_dir")
    if not language:
        language = get_lang_dirs(locale_dir)
    languages = sum(language, ())  # flatten
    if not languages:
        msg = (
            "No languages are found. Please specify language with -l "
            "option, or preparing language directories under %(locale_dir)r "
            "directory." % locals()
        )
        raise click.BadParameter(msg, param_hint="language")

    basic.update(
        locale_dir,
        pot_dir,
        languages,
        line_width,
        ignore_obsolete=no_obsolete,
        jobs=jobs,
    )


@main.command()
@option_locale_dir
@option_output_dir
@option_language
def build(locale_dir, output_dir, language):
    """
    Build specified language's po files into mo.
    """
    if not language:
        language = get_lang_dirs(locale_dir)
    languages = sum(language, ())  # flatten

    if not output_dir or (
        os.path.exists(output_dir) and os.path.samefile(locale_dir, output_dir)
    ):
        output_dir = locale_dir

    basic.build(locale_dir, output_dir, languages)


@main.command()
@option_locale_dir
@option_language
def stat(locale_dir, language):
    """
    Print statistics for all po files.
    """
    if not language:
        language = get_lang_dirs(locale_dir)
    languages = sum(language, ())  # flatten
    basic.stat(locale_dir, languages)


@main.command("create-transifexrc")
@option_transifex_token
def create_transifexrc(transifex_token):
    """
    Create `$HOME/.transifexrc`  (deprecated)
    """
    warnings.filterwarnings("default", category=DeprecationWarning)
    warnings.warn(
        (
            "create-transifexrc command and SPHINXINTL_TRANSIFEX_TOKEN "
            "will be deprecated. Please use TX_TOKEN environment variable instead."
        ),
        category=DeprecationWarning,
    )
    transifex.create_transifexrc(transifex_token)


@main.command("create-txconfig")
def create_txconfig():
    """
    Create `./.tx/config`
    """
    transifex.create_txconfig()


@main.command("update-txconfig-resources")
@option_transifex_organization_name
@option_transifex_project_name
@option_locale_dir
@option_pot_dir
def update_txconfig_resources(
    transifex_organization_name, transifex_project_name, locale_dir, pot_dir
):
    """
    Update resource sections of `./.tx/config`.
    """
    if not pot_dir:
        pot_dir = os.path.join(locale_dir, "pot")

    transifex.update_txconfig_resources(
        transifex_organization_name, transifex_project_name, locale_dir, pot_dir
    )


if __name__ == "__main__":
    main(auto_envvar_prefix=ENVVAR_PREFIX)
